
// TODO
// - Lookup Table implementieren -> Weights, roots schon früher rausziehen? 
// -> Warum macht delta_lut alles langsamer?
// - Omnes auch komplex integrieren (geht nicht mit Integralmethode von gauss-quad)
// - Textfelder und "Berechnen"-Button in UI, um z.B. Wertebereich zu setzen
// - Neue Architketur: struct calculator, das lookups speichert -> In die nötigen Funktionen übergeben. Dadurch kann man omnes auch als single calc machen und trotzdem die Punkte nur einmal berechnen
// - AutoImport ausschalten


use egui_plot::{log_grid_spacer, AxisHints, GridInput, GridMark, Legend, Line, Plot};
use num::complex::ComplexFloat;
use std::time::Instant;

mod gq_storage;
mod calc;

struct App {
    plot_data: Vec<PlotCurve>,
    plots_available: [String; 6],
    calc_cache: calc::Cache,
    scaling_type: ScalingType, 
    a: f64,
    b: f64, 
    n_points: u32,
    n_omnes: u32,
    n_phi0: u32,
    use_tan_omnes: bool,
    use_tan_phi0: bool,
    use_xsub_phi0: bool
}

#[derive(Clone)]
struct PlotCurve {
    points: Vec<[f64;2]>,
    name: String,
}

impl Default for App {
    fn default() -> Self {
        App {
            plot_data: Vec::default(),
            plots_available: ["Delta".to_string(), "Omnes Integrand".to_string(), "Omnes".to_string(), "Phi0 Integrand".to_string(), "Phi0".to_string(), "Phi0 / delta".to_string()],
            calc_cache: calc::Cache::default(),
            scaling_type: ScalingType::default(),
            a: calc::S0,
            b: 9.0,
            n_points: 5000,
            n_omnes:1000,
            n_phi0:1000,
            use_tan_omnes: true,
            use_tan_phi0: true,
            use_xsub_phi0: true
        }
    }
}

#[derive(Debug, Default, PartialEq, Clone, Copy)]
enum ScalingType {
    Logarithmic, 
    #[default]
    Linear
}

impl eframe::App for App {
    fn update(&mut self, ctx: &egui::Context, _frame: &mut eframe::Frame) {
        egui::CentralPanel::default().show(ctx, |ui| {
            let cur_scaling_type = self.scaling_type;

            ui.vertical(|ui| {
                ui.horizontal(|ui| {
                    egui::Grid::new("grid1")
                        .min_col_width(125.0)
                        .show(ui, |ui| {
                        ui.label("calc interval start:").on_hover_text("the upper bound of the interval that functions are calculated in");;
                        float_text_edit_singleline(ui, &mut self.a);
                        ui.end_row();

                        ui.label("calc interval end:").on_hover_text("the upper bound of the interval that functions are calculated in");
                        float_text_edit_singleline(ui, &mut self.b);
                        ui.end_row();
                    
                        ui.label("N for omnes:").on_hover_text("the number of points used in calculation of omnes function");
                        int_text_edit_singleline(ui, &mut self.n_omnes);
                        ui.end_row();

                        ui.label("N for phi0:").on_hover_text("the number of points used in calculation of phi0 function");
                        int_text_edit_singleline(ui, &mut self.n_phi0);
                        ui.end_row();

                        ui.label("number of plotpoints:").on_hover_text("the number of points shown in the plot");
                        int_text_edit_singleline(ui, &mut self.n_points);
                        ui.end_row();
                    });

                    ui.separator();

                    ui.vertical(|ui| {
                        ui.checkbox(&mut self.use_tan_omnes, "Use tan()-subst. for Omnes");
                        ui.checkbox(&mut self.use_tan_phi0, "Use tan()-subst. for phi0");
                        ui.checkbox(&mut self.use_xsub_phi0, "Use second subst. for phi0");
                        egui::ComboBox::from_label("pick y-scaling").selected_text(format!("{:?}", self.scaling_type)).show_ui(ui, |ui| {
                            ui.selectable_value(&mut self.scaling_type, ScalingType::Logarithmic, "log10");
                            ui.selectable_value(&mut self.scaling_type, ScalingType::Linear, "linear");
                        });
                        
                        ui.horizontal(|ui| {
                            let button_clear = ui.button("Clear canvas");
                            if button_clear.clicked() {
                                self.plot_data = Vec::new();
                            }
                            egui::widgets::global_dark_light_mode_switch(ui);
                        });
                        
                    });
                });
            
                ui.separator(); 
            
                ui.label("You can calculate the following functions:");

                ui.horizontal( |ui| {
                    for i in 0..self.plots_available.len() {
                        let button = ui.button( format!("{}", self.plots_available[i].clone()));
                        if button.clicked() {
                            let x_values: Vec<f64> = (0..self.n_points).map(|x| -> f64 {f64::from(x) * ((self.b - self.a) / f64::from(self.n_points)) + self.a}).collect();
                            match i {
                                0 => {
                                    let y_values_delta:Vec<f64> = if self.scaling_type == ScalingType::Linear {
                                        x_values.iter().map(|&x| calc::delta_with_lut(&mut self.calc_cache, x)).collect()
                                    } else {
                                        x_values.iter().map(|&x| calc::delta_with_lut(&mut self.calc_cache, x).log10()).collect()
                                    };
                                    
                                    self.plot_data.push(PlotCurve {
                                        points: x_values.iter().zip(y_values_delta.iter()).map(|(&x,&y)| [x.powf(0.5),y]).collect(),
                                        name: "Delta".to_string()
                                    })
                                }
                                1 => {
                                    let y_values_omnes_integrand:Vec<f64> = if self.scaling_type == ScalingType::Linear {
                                        x_values.iter().map(|&x| calc::omnes_integrand_tan(&mut self.calc_cache, x, 100.0).abs()).collect()
                                    } else {
                                        x_values.iter().map(|&x| calc::omnes_integrand_tan(&mut self.calc_cache, x, 100.0).abs().log10()).collect()
                                    };
                                    self.plot_data.push(PlotCurve {
                                        points: x_values.iter().zip(y_values_omnes_integrand.iter()).map(|(&x,&y)| [x.powf(0.5),y]).collect(),
                                        name: "Omnes Integrand".to_string()
                                    })
                                }
                                2 => {
                                    let y_values_omnes:Vec<f64> = if self.scaling_type == ScalingType::Linear {
                                        x_values.iter().map(|&x| calc::omnes(&mut self.calc_cache, x, self.n_omnes, self.use_tan_omnes).abs().powi(2)).collect()
                                    } else {
                                        x_values.iter().map(|&x| calc::omnes(&mut self.calc_cache, x, self.n_omnes, self.use_tan_omnes).abs().powi(2).log10()).collect()
                                    };
                                    self.plot_data.push(PlotCurve {
                                        points: x_values.iter().zip(y_values_omnes.iter()).map(|(&x,&y)| [x.powf(0.5),y]).collect(),
                                        name: "Omnes".to_string()
                                    })
                                }
                                3 => {
                                    let y_values_phi0_integrand:Vec<f64> = if self.scaling_type == ScalingType::Linear {
                                        x_values.iter().map(|&x| calc::phi0_integrand_tan_xsub(&mut self.calc_cache, x, 10000.0, self.n_omnes, self.use_tan_omnes)).collect()
                                    } else {
                                        x_values.iter().map(|&x| calc::phi0_integrand_tan_xsub(&mut self.calc_cache, x, 10000.0, self.n_omnes, self.use_tan_omnes).log10()).collect()
                                    };
                                    self.plot_data.push(PlotCurve {
                                        points: x_values.iter().zip(y_values_phi0_integrand.iter()).map(|(&x,&y)| [x.powf(0.5),y]).collect(),
                                        name: "Phi0 Integrand".to_string()
                                    })
                                }
                                4 => {
                                    let y_values_phi0: Vec<f64> = if self.scaling_type == ScalingType::Linear {
                                        x_values.iter().map(|&x| calc::phi0(&mut self.calc_cache, x, self.n_omnes, self.n_phi0, self.use_tan_omnes, self.use_tan_phi0, self.use_xsub_phi0)).collect()
                                    } else {
                                        x_values.iter().map(|&x| calc::phi0(&mut self.calc_cache, x, self.n_omnes, self.n_phi0, self.use_tan_omnes, self.use_tan_phi0, self.use_xsub_phi0).log10()).collect()
                                    };
                                    self.plot_data.push(PlotCurve {
                                        points: x_values.iter().zip(y_values_phi0.iter()).map(|(&x,&y)| [x.powf(0.5),y]).collect(),
                                        name: "Phi0".to_string()
                                    })
                                    // let y_values_phi0 = calc::bit_pattern_randomness_tester(self.a, self.b, self.n_points as u64);
                                    // self.plot_data.push(PlotCurve {
                                    //     points: (0..self.n_points).zip(y_values_phi0.iter()).map(|(x,&y)| [x as f64,y]).collect(),
                                    //     name: "Phi0".to_string()
                                    // })
                                }
                                5 => {
                                    let y_values_phi0: Vec<f64> = if self.scaling_type == ScalingType::Linear {
                                        x_values.iter().map(|&x| {
                                            let val = calc::phi0(&mut self.calc_cache, x.powi(2), self.n_omnes, self.n_phi0, self.use_tan_omnes, self.use_tan_phi0, self.use_xsub_phi0) / calc::delta_with_lut(&mut self.calc_cache, x.powi(2));
                                            val
                                        }).collect()
                                    } else {
                                        x_values.iter().map(|&x| {
                                            let val = calc::phi0(&mut self.calc_cache, x.powi(2), self.n_omnes, self.n_phi0, self.use_tan_omnes, self.use_tan_phi0, self.use_xsub_phi0) / calc::delta_with_lut(&mut self.calc_cache, x.powi(2));
                                            val.log10()
                                        }).collect()
                                    };
                                    self.plot_data.push(PlotCurve {
                                        points: x_values.iter().zip(y_values_phi0.iter()).map(|(&x,&y)| [x,y]).collect(),
                                        name: "phi0 / delta".to_string()
                                    })                            
                                }
                                6_usize.. => {
                                    panic!("Not all buttons have been assigned data!");
                                }
                            }
                        }
                    }
                });
            });
            
            if self.scaling_type != cur_scaling_type {
                for curve in self.plot_data.iter_mut() {    
                    for [_x,y] in curve.points.iter_mut() {
                        if self.scaling_type == ScalingType::Linear {
                            *y = 10.0f64.powf(*y);
                        } else if self.scaling_type == ScalingType::Logarithmic {
                            *y = y.log10();
                        }
                    }
                }
            }

            let plot = Plot::new("my_plot")
                .grid_spacing(5.0f32..=300.0f32)
                .x_axis_label("sqrt(s) [GeV]")
                .legend(Legend::default());

            let plot = match self.scaling_type {
                ScalingType::Linear => plot.custom_y_axes(vec![AxisHints::new_y().formatter(|mark, _l, _| scientific_notation(mark.value, false))]),
                ScalingType::Logarithmic => {plot.y_grid_spacer(|input| {
                        let input_exp = GridInput {bounds: (inverse_log_map(input.bounds.0), inverse_log_map(input.bounds.1)), base_step_size: input.base_step_size};
                        let lin_gridmarks = (log_grid_spacer(10))(input_exp);
                        let mut log_gridmarks: Vec<GridMark> = Vec::new();
                        for v in lin_gridmarks.iter() {
                            if v.value - v.value.floor() < 0.9 {
                                log_gridmarks.push(GridMark{value:log_map(v.value), step_size:v.step_size / 1.14}); 
                            }        
                        }
                        log_gridmarks
                    })
                    .label_formatter(|_,point| format!("x = {:.5}\ny = {:.5}", point.x, 10.0f64.powf(point.y)).to_string())
                    .custom_y_axes(vec![AxisHints::new_y().formatter(|mark, _l, _| scientific_notation(mark.value, true))])
                }
            };     
        
            
            plot.show(ui, |plot_ui| {
                for i in 0..self.plot_data.len() {
                    plot_ui.line(Line::new(self.plot_data[i].points.clone()).name(self.plot_data[i].name.clone())); // is clone necessary?
                }
            });
        });
    }
}

#[cfg(not(target_arch = "wasm32"))]
fn main() {
    let mut app = App::default();
    app.calc_cache = calc::Cache::from_file("./gauss_quad_lut.morello").unwrap();
    // let x_values: Vec<f64> = (0..app.n_points).map(|x| -> f64 {f64::from(x) * ((app.b - app.a) / f64::from(app.n_points)) + app.a}).collect();
    // let t0 = Instant::now();
    // let y_values_phi0: Vec<f64> = x_values.iter().map(|&x| calc::phi0(&mut app.calc_cache, x, app.n_omnes, app.n_phi0)).collect();
    // std::hint::black_box(&y_values_phi0);
    // println!("{:?}", t0.elapsed());

    
    eframe::run_native("Morellus", eframe::NativeOptions::default(), Box::new(|_cc| Box::new(app))).unwrap();
}

#[cfg(target_arch = "wasm32")]
fn main() {
    let mut app = App::default();

    wasm_bindgen_futures::spawn_local(async {
        let res = gloo_net::http::RequestBuilder::new("/lut/gauss_quad_lut.morello")
        .method(gloo_net::http::Method::POST)
        .send()
        .await
        .unwrap();

        app.calc_cache = calc::Cache::from_slice(&res.binary().await.unwrap()).unwrap();
        let web_options = eframe::WebOptions::default();

        eframe::WebRunner::new()
            .start(
                "morellus_canvas_id",
                web_options,
                Box::new(|_cc| Box::new(app)),
            )
            .await
            .expect("failed to start eframe");
    });
}

fn log_map(x:f64) -> f64 {
    x.floor() + (1.0 + (x - x.floor()) * 10.0).log10()
}

fn inverse_log_map(x:f64) -> f64 {
    x.floor() + (10f64.powf(x - x.floor()) - 1.0) / 10.0
}

fn scientific_notation(x:f64, log:bool) -> String { 
    if log {
        format!("{}*10^{}", format!(
            "{:.8}", 10f64.powf(x - x.floor()))
            .trim_end_matches('0').trim_end_matches('.').to_string(),
            x.floor()).to_string()
    } else {
        format!("{:.9}", x).trim_end_matches('0').trim_end_matches('.').to_string()
    }
}

fn int_text_edit_singleline(ui: &mut egui::Ui, value: &mut u32) -> egui::Response {
    let mut tmp_value = format!("{}", value);
    let res = ui.text_edit_singleline(&mut tmp_value);
    if let Ok(result) = tmp_value.parse::<u32>() {
        *value = result;
    } else if tmp_value == "" {
        *value = 0;
        
    }
    res
}

fn float_text_edit_singleline(ui: &mut egui::Ui, value: &mut f64) -> egui::Response {
    let mut tmp_value = format!("{}", value);
    let res = ui.text_edit_singleline(&mut tmp_value);
    if let Ok(result) = tmp_value.parse::<f64>() {
        *value = result;
    } else if tmp_value == "" {
        *value = 0.0;
    }
    res
}